from .base import BaseView, Resource
from .proxy import ConfProxyView, ProxyView, TransparentProxyView
from .resource import CatalogResource, Restful
from .rpc import JSONRPCView
from .template import RenderTemplate



import os
import sys

import jinja2
from werkzeug.middleware.lint import LintMiddleware
from werkzeug.middleware.profiler import ProfilerMiddleware

from . import config, flaskel
from .converters import CONVERTERS
from .utils import misc
from .utils.datastruct import ObjectDict


class AppBuilder:
    """Flask app builder"""

    """

    default app name
    """
    app_name = config.FLASK_APP

    """
    default secret key file name
    """
    secret_file = '.secret.key'

    """
    custom url converters
    """
    url_converters = CONVERTERS

    """
    custom app config module
    """
    conf_module = config

    def __init__(self, conf_module=None, extensions=None,
                 converters=(), blueprints=(), folders=(), middlewares=(), **options):
        """

        :param conf_module: python module file
        :param converters: custom url converter mapping
        :param extensions: custom extension appended to defaults
        :param blueprints: list of application's blueprints
        :param middlewares: list of wsgi middleware
        :param folders: list of custom jinja2's template folders
        :param options: passed to Flask class
        :return:
        """
        self._app = None
        self._converters = converters
        self._blueprints = blueprints
        self._extensions = extensions
        self._middlewares = middlewares
        self._folders = folders
        self._options = options
        self._conf_module = conf_module or self.conf_module

    def _generate_secret_key(self, secret_file, key_length):
        """

        :param secret_file:
        :param key_length:
        """
        secret_key = misc.random_string(key_length)

        with open(secret_file, 'w') as f:
            f.write(secret_key)
            abs_file = os.path.abspath(secret_file)
            self._app.logger.warning(f"new secret key generated: take care of this file:{abs_file}")
        return secret_key

    def _load_secret_key(self, secret_file):
        """

        :param secret_file:
        :return:
        """
        if not os.path.isfile(secret_file):
            return None

        with open(secret_file, 'r') as f:
            secret_key = f.read()
            self._app.logger.info(f"load secret key from: {secret_file}")
        return secret_key

    def _set_secret_key(self):
        """

        """
        secret_key = None
        key_length = self._app.config.SECRET_KEY_MIN_LENGTH

        if not self._app.config.get('FLASK_ENV', '').lower().startswith('dev'):
            secret_file = self._app.config.SECRET_KEY

            if secret_file:
                secret_key = self._load_secret_key(secret_file) or secret_file
            else:
                secret_file = self.secret_file
                secret_key = self._load_secret_key(secret_file)
                secret_key = secret_key or self._generate_secret_key(secret_file, key_length)
        elif not self._app.config.get('SECRET_KEY'):
            self._app.logger.debug('set secret key in development mode')
            secret_key = 'fake_very_complex_string'

        self._app.config.SECRET_KEY = secret_key or self._app.config.SECRET_KEY
        if len(secret_key) < key_length:
            self._app.logger.warning(f"secret key length is less than: {key_length}")

    def _register_extensions(self):
        """

        """
        with self._app.app_context():
            for name, e in (self._extensions or {}).items():
                try:
                    ext = e[0]
                    opt = e[1] if len(e) > 1 else {}
                    if not ext:
                        raise TypeError("extension could not be None")
                except (TypeError, IndexError) as exc:
                    self._app.logger.debug(
                        f"Invalid extension '{name}' configuration '{e}': {exc}"
                    )
                    continue

                ext.init_app(self._app, **opt)
                self._app.logger.debug(
                    f"Registered extension '{name}' with options: {str(opt)}"
                )

    def _register_blueprints(self):
        """

        """
        for b in (self._blueprints or []):
            try:
                bp = b[0]
                opt = b[1] if len(b) > 1 else {}
                if not bp:
                    raise TypeError('blueprint could not be None')
            except (TypeError, IndexError) as exc:
                self._app.logger.debug(f"invalid blueprint configuration '{b}': {exc}")
                continue

            self._app.register_blueprint(bp, **opt)
            self._app.logger.debug(f"Registered blueprint '{bp.name}' with options: {str(opt)}")

    def _set_config(self, conf):
        """

        :param conf:
        :return:
        """
        self._app.config.from_object(self._conf_module)
        self._app.config.from_mapping(**(conf or {}))
        self._app.config.from_envvar('APP_CONFIG_FILE', silent=True)
        self._app.config = ObjectDict(**self._app.config)

    def _register_converters(self):
        """

        """
        conv = {**self.url_converters, **(self._converters or {})}
        self._app.url_map.converters.update(conv)

        for k in conv.keys():
            self._app.logger.debug(f"Registered converter: '{k}'")

    def _register_template_folders(self):
        """

        """
        loaders = [self._app.jinja_loader]

        for fsl in self._folders:
            loaders.append(jinja2.FileSystemLoader(fsl))

        if self._folders:
            folders = ", ".join(self._folders)
            self._app.jinja_loader = jinja2.ChoiceLoader(loaders)
            self._app.logger.debug(f"Registered template folders: '{folders}'")

    def _register_middlewares(self):
        """

        """
        for middleware in (self._middlewares or []):
            # WorkAround: in order to pass flask app to middleware without breaking chain
            if not (hasattr(middleware, 'flask_app') and middleware.flask_app):
                setattr(middleware, 'flask_app', self._app)

            self._app.wsgi_app = middleware(self._app.wsgi_app)
            self._app.logger.debug(f"Registered middleware: '{middleware}'")

    def _set_linter_and_profiler(self):
        if self._app.config.WSGI_WERKZEUG_LINT_ENABLED:
            self._app.wsgi_app = LintMiddleware(self._app.wsgi_app)

        if self._app.config.WSGI_WERKZEUG_PROFILER_ENABLED:
            stream = self._app.config.WSGI_WERKZEUG_PROFILER_FILE
            stream = open(stream, 'w') if stream else sys.stdout
        else:
            stream = None  # pragma: no cover

        self._app.wsgi_app = ProfilerMiddleware(
            self._app.wsgi_app, stream=stream,
            restrictions=self._app.config.WSGI_WERKZEUG_PROFILER_RESTRICTION
        )

    def _dump_urls(self):
        output = []
        for rule in self._app.url_map.iter_rules():
            methods = ','.join(rule.methods)
            output.append("{:30s} {:40s} {}".format(rule.endpoint, methods, rule))

        lines = "\n".join(sorted(output))
        self._app.logger.debug(f"Registered routes:\n{lines}")

    def _patch_app(self):
        if self._app.debug:
            self._set_linter_and_profiler()
            self._dump_urls()

        @self._app.before_first_request
        def create_database():
            sqlalchemy = self._app.extensions.get('sqlalchemy')
            if sqlalchemy is not None:
                sqlalchemy.db.create_all()

    def create(self, conf=None):
        """

        :param conf:
        :return:
        """
        self._app = flaskel.Flaskel(self.app_name, **self._options)

        self._set_config(conf)
        self._set_secret_key()

        self._register_extensions()
        self._register_middlewares()
        self._register_template_folders()
        self._register_converters()
        self._register_blueprints()

        self._patch_app()

    def get_or_create(self, conf=None):
        """

        :param conf:
        :return:
        """
        if self._app is None:
            self.create(conf)

        return self._app



from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.wsgi import WSGIContainer

from .base import BaseApplication


class WSGITornado(BaseApplication):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        BaseApplication.__init__(self, app, options)
        self._http_server = HTTPServer(WSGIContainer(self.application))

    def run(self):
        """

        """
        self._http_server.listen(address=self._interface, port=self._port)
        IOLoop.current().start()



# noinspection PyUnresolvedReferences
from flaskel.tester import Asserter, HttpAsserter
# noinspection PyUnresolvedReferences
from . import testapp


def test_app_runs(testapp):
    res = testapp.get('/')
    HttpAsserter.assert_status_code(res)



import flask

from .base import BaseView


class RenderTemplate(BaseView):
    methods = ['GET', 'POST']

    def __init__(self, template, content_type=None, **kwargs):
        """

        :param template:
        """
        self._params = kwargs
        self._template = template
        self._content_type = content_type

    # noinspection PyUnusedLocal
    def service(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return kwargs or self._params

    def response(self, data, **kwargs):
        return flask.Response(data, mimetype=self._content_type, **kwargs)

    def dispatch_request(self):
        return self.response(
            flask.render_template(self._template, **self.service())
        )


class RenderTemplateString(RenderTemplate):
    def dispatch_request(self):
        return self.response(
            flask.render_template_string(self._template, **self.service())
        )



from flaskel.http import rpc


# noinspection PyPep8Naming,PyMethodMayBeStatic,PyUnusedLocal
class MyJsonRPC:
    def testAction1(self, **kwargs):
        return {
            "action1": True
        }

    def testAction2(self, **kwargs):
        return {
            "action2": True
        }

    def testInvalidParams(self, **kwargs):
        if kwargs.get('param') != 'testparam':
            raise rpc.RPCInvalidParams()

    def testInternalError(self):
        raise ValueError



import flask
from requests import auth, exceptions as http_exc, request as send_request

from flaskel import cap
from flaskel.utils.datastruct import ObjectDict
from flaskel.utils.faker import FakeLogger
from flaskel.utils.uuid import get_uuid
from . import httpcode
from .httpdumper import FlaskelHTTPDumper, HTTPDumper


class HTTPTokenAuth(auth.AuthBase):
    def __init__(self, token):
        """

        :param token:
        """
        self.token = token

    def __eq__(self, other):
        """

        :param other:
        :return:
        """
        return self.token == getattr(other, 'token', None)

    def __ne__(self, other):
        """

        :param other:
        :return:
        """
        return not self == other

    def __call__(self, r):
        """

        :param r:
        :return:
        """
        r.headers['Authorization'] = f"Bearer {self.token}"
        return r


class HTTPBase(HTTPDumper):
    def __init__(self, dump_body=False, raise_on_exc=False, logger=None):
        """

        :param dump_body:
        :param raise_on_exc:
        :param logger:
        """
        self._dump_body = dump_body
        self._raise_on_exc = raise_on_exc
        self._logger = logger or FakeLogger()

    def request(self, uri, dump_body=None, **kwargs):
        """

        :param uri:
        :param dump_body:
        :param kwargs:
        """
        raise NotImplemented  # pragma: no cover


class HTTPClient(HTTPBase):
    def __init__(self, endpoint, token=None, username=None, password=None, **kwargs):
        """

        :param endpoint:
        :param token:
        :param username:
        :param password:
        """
        super().__init__(**kwargs)
        self._endpoint = endpoint
        self._username = username
        self._password = password
        self._token = token

    def get_auth(self):
        """

        :return:
        """
        if self._username and self._password:
            return auth.HTTPBasicAuth(self._username, self._password)

        if self._token:
            return HTTPTokenAuth(self._token)

    def normalize_url(self, url):
        if url.startswith('http'):
            return url

        return f"{self._endpoint}/{url.rstrip('/')}"

    def request(self, uri, method='GET', raise_on_exc=False,
                dump_body=None, chunk_size=None, decode_unicode=False, **kwargs):
        """

        :param uri:
        :param method:
        :param raise_on_exc:
        :param dump_body:
        :param chunk_size:
        :param decode_unicode:
        :param kwargs:
        :return:
        """
        kwargs['auth'] = self.get_auth()
        if dump_body is None:
            dump_body = self._dump_body
        if kwargs.get('stream') is True:
            dump_body = False

        try:
            response = send_request(method, self.normalize_url(uri), **kwargs)
            self._logger.info(self.dump_request(response.request, dump_body))
        except (http_exc.ConnectionError, http_exc.Timeout) as exc:
            self._logger.exception(exc)
            if raise_on_exc or self._raise_on_exc:
                raise  # pragma: no cover

            return ObjectDict(
                body={},
                status=httpcode.SERVICE_UNAVAILABLE,
                headers={},
                exception=exc
            )

        try:
            response.raise_for_status()
        except http_exc.HTTPError as exc:
            self._logger.warning(self.dump_response(response, dump_body))
            response = exc.response
            if raise_on_exc or self._raise_on_exc:
                raise
        else:
            self._logger.info(self.dump_response(response, dump_body))

        if kwargs.get('stream') is True:
            body = response.iter_content(chunk_size, decode_unicode)
        elif 'json' in (response.headers.get('Content-Type') or ''):
            body = response.json()
        else:
            body = response.text

        return ObjectDict(
            body=body,
            status=response.status_code,
            headers=dict(response.headers)
        )

    def get(self, uri, **kwargs):
        return self.request(uri, **kwargs)

    def post(self, uri, **kwargs):
        return self.request(uri, method='POST', **kwargs)

    def put(self, uri, **kwargs):
        return self.request(uri, method='PUT', **kwargs)

    def patch(self, uri, **kwargs):
        return self.request(uri, method='PATCH', **kwargs)

    def delete(self, uri, **kwargs):
        return self.request(uri, method='DELETE', **kwargs)

    def options(self, uri, **kwargs):
        return self.request(uri, method='OPTIONS', **kwargs)

    def head(self, uri, **kwargs):
        return self.request(uri, method='HEAD', **kwargs)


class JsonRPCClient(HTTPClient):
    def __init__(self, endpoint, uri, version='2.0', **kwargs):
        """

        :param endpoint:
        :param uri:
        :param version:
        """
        super().__init__(endpoint, raise_on_exc=True, **kwargs)
        self._uri = uri
        self._version = version
        self._request_id = None

    @property
    def request_id(self):
        """

        :return:
        """
        return self._request_id

    def request(self, method, request_id=None, **kwargs):
        """

        :param method:
        :param request_id:
        :param kwargs:
        :return:
        """
        self._request_id = request_id or get_uuid()
        return self._request(method, **kwargs)

    def notification(self, method, **kwargs):
        """

        :param method:
        :param kwargs:
        :return:
        """
        self._request_id = None
        return self._request(method, **kwargs)

    def _request(self, method, params=None, **kwargs):
        """

        :param method:
        :param params:
        :param kwargs:
        :return:
        """
        resp = super().request(
            self._uri,
            method='POST',
            json=dict(
                jsonrpc=self._version,
                method=method,
                params=params or {},
                id=self._request_id
            ),
            **kwargs
        )

        return resp.body or ObjectDict()


class FlaskelHttp(HTTPClient, FlaskelHTTPDumper):
    def __init__(self, endpoint, **kwargs):
        kwargs.setdefault('logger', cap.logger)
        super().__init__(endpoint, **kwargs)

    def request(self, uri, **kwargs):
        if flask.request.id:
            if not kwargs.get('headers'):
                kwargs['headers'] = {}
            kwargs['headers'][cap.config.REQUEST_ID_HEADER] = flask.request.id

        return super().request(uri, **kwargs)



# example spa blueprint
#
import os

from flask import Blueprint

from flaskel import cap
from flaskel.ext import errors

static_folder = os.environ.get('SPA_STATIC_FOLDER') or 'webapp'
static_url_path = os.environ.get('SPA_STATIC_URL_PATH') or '/static/'

spa = Blueprint(
    'spa', __name__,
    static_folder=static_folder,
    static_url_path=static_url_path
)

errors.web_register(spa)


@spa.route('/', defaults=dict(path=''))
@spa.route('/<path:path>')
def catch_all(path):
    """

    :param path:
    :return:
    """
    return cap.send_static_file("index.html")



import socket

from flask_mail import Message

from flaskel import cap
from . import celery


@celery.task
def send_email(email_data):
    """

    :param email_data:
    """
    message = Message(
        email_data['subject'],
        sender=cap.config['MAIL_DEFAULT_SENDER'],
        recipients=[
            email_data['to']
        ]
    )
    message.body = email_data['body']
    mail = cap.extensions['mail']

    old_timeout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(60)

    mail.send(message)
    socket.setdefaulttimeout(old_timeout)



import flask

from flaskel import cap
from flaskel.http import FlaskelHttp
from flaskel.utils.datastruct import ObjectDict
from .base import BaseView


class ProxyView(BaseView):
    def __init__(self, host=None, url=None, method=None,
                 proxy_body=False, proxy_headers=False, proxy_params=False):
        """

        :param host:
        :param url:
        :param method:
        :param proxy_body:
        :param proxy_headers:
        :param proxy_params:
        """
        self._host = host
        self._method = method
        self._url = url
        self._proxy_body = proxy_body
        self._proxy_headers = proxy_headers
        self._proxy_params = proxy_params

    def dispatch_request(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        response = self.proxy(self.service())
        return flask.Response(
            flask.stream_with_context(response.body),
            headers=response.headers,
            status=response.status
        )

    def proxy(self, data):
        return FlaskelHttp(data.host or self.upstream_host()).request(
            data.url or self.request_url(),
            method=data.method or self.request_method(),
            headers=data.headers or self.request_headers(),
            params=data.params or self.request_params(),
            data=data.body or self.request_body(),
            stream=True
        )

    def service(self):
        """

        :return: ObjectDict
        """
        return ObjectDict(
            host=self.upstream_host(),
            uri=self.request_url(),
            method=self.request_method(),
            headers=self.request_headers(),
            params=self.request_params(),
            data=self.request_body()
        )

    def upstream_host(self):
        return self._host

    def request_url(self):
        return self._url or flask.request.path

    def request_method(self):
        return self._method or flask.request.method

    def request_body(self):
        if self._proxy_body:
            return flask.request.get_data()

    def request_headers(self):
        if self._proxy_headers:
            return flask.request.headers

    def request_params(self):
        if self._proxy_params:
            return flask.request.params


class ConfProxyView(ProxyView):
    def __init__(self, config_key, **kwargs):
        super().__init__(**kwargs)
        self._config_key = config_key

    def service(self):
        """

        :return: ObjectDict
        """
        conf = cap.config
        keys = self._config_key.split('.')
        for k in keys:
            conf = conf.get(k) or ObjectDict()

        return ObjectDict(**conf)


class TransparentProxyView(ProxyView):
    def __init__(self, **kwargs):
        kwargs.setdefault('proxy_body', True)
        kwargs.setdefault('proxy_headers', True)
        kwargs.setdefault('proxy_params', True)
        super().__init__(**kwargs)



import functools

from flaskel import cap, httpcode
from flaskel.ext import builder


class HealthCheck:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        """
        self._health_checks = {}

        if app:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, bp=None, extensions=()):
        """

        :param app:
        :param bp:
        :param extensions:
        """
        app.config.setdefault('HEALTHCHECK_ABOUT_LINK', None)
        app.config.setdefault('HEALTHCHECK_PATH', '/healthcheck')
        app.config.setdefault('HEALTHCHECK_CONTENT_TYPE', 'application/health+json')

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['healthcheck'] = self

        blueprint = bp or app
        blueprint.add_url_rule(
            app.config['HEALTHCHECK_PATH'],
            view_func=self.do_health_check
        )

        for ex in extensions:
            self.register(**ex)(ex.pop('func'))

    @builder.response('json')
    def do_health_check(self):
        """

        :return:
        """
        healthy = True
        response = dict(
            status='pass',
            checks={},
            links={"about": cap.config['HEALTHCHECK_ABOUT_LINK']}
        )

        for name, check_func in self._health_checks.items():
            try:
                state, message = check_func()
            except Exception as exc:  # pragma: no cover
                state = False
                message = str(exc)

            healthy = state
            response['checks'][name] = dict(
                status='pass' if state else 'fail',
                output=message
            )

        if not healthy:
            response['status'] = 'fail'

        status = httpcode.SUCCESS if healthy else httpcode.SERVICE_UNAVAILABLE
        return response, status, {'Content-Type': cap.config['HEALTHCHECK_CONTENT_TYPE']}

    def register(self, name=None, **kwargs):
        """

        :param name:
        :param kwargs:
        :return:
        """

        def _register(func):
            """

            :param func:
            :return:
            """
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                self._health_checks[name] = functools.partial(func, **kwargs)

            return wrapped()

        return _register



# example auth blueprint
#
from flask import Blueprint

from flaskel.ext import cors, errors

auth = Blueprint('auth', __name__, url_prefix='/auth')

cors.init_app(auth)
errors.api_register(auth)

from . import token



__version__ = '1.3.0rc1'

__author_name__ = 'cs91chris'
__author_email__ = 'cs91chris@voidbrain.me'

__author__ = f"{__author_name__} <{__author_email__}>"

__all__ = [
    '__version__',
    '__author__',
    '__author_name__',
    '__author_email__',
]



from flask import Blueprint

from flaskel.ext import errors
from flaskel.views import RenderTemplate

web = Blueprint(
    'web', __name__,
    template_folder="templates",
    static_folder="static"
)

errors.web_register(web)
RenderTemplate.register(web, url='/', name='index', template='index.html')



from blueprints import BLUEPRINTS
from ext import EXTENSIONS
from flaskel import AppBuilder, BASE_EXTENSIONS, middlewares as middle, Server

factory = AppBuilder(
    blueprints=BLUEPRINTS,
    extensions={**BASE_EXTENSIONS, **EXTENSIONS},
    middlewares=(
        middle.RequestID,
        middle.HTTPMethodOverride,
        middle.ReverseProxied,
    )
)


def create_app(conf=None):
    """
    useful for external wsgi server

    :param conf:
    :return:
    """
    return factory.get_or_create(conf)


def cli():
    Server(factory).run_from_cli()


if __name__ == '__main__':
    cli()



import asyncio
import warnings

import flask

from flaskel import cap
from flaskel.utils.datastruct import ObjectDict
from flaskel.utils.uuid import get_uuid
from .client import HTTPBase, httpcode
from .httpdumper import FlaskelHTTPDumper

try:
    import aiohttp
except ImportError as err:  # pragma: no cover
    warnings.warn(str(err))
    aiohttp = None

try:
    import nest_asyncio
except ImportError as err:  # pragma: no cover
    warnings.warn(str(err))
    nest_asyncio = None


class HTTPBatch(HTTPBase):
    def __init__(self, conn_timeout=60, read_timeout=60, **kwargs):
        """

        :param conn_timeout:
        :param read_timeout:
        :param kwargs:
        """
        super().__init__(**kwargs)
        self._timeout = aiohttp.ClientTimeout(
            sock_read=read_timeout,
            sock_connect=conn_timeout
        )

    async def http_request(self, session, **kwargs):
        """

        :param session:
        :param kwargs:
        :return:
        """
        if not aiohttp:
            raise ImportError("You must install 'aiohttp'")  # pragma: no cover
        try:
            async with session.request(**kwargs) as resp:
                # noinspection PyProtectedMember
                self._logger.info(self.dump_request(resp._request_info, self._dump_body))
                try:
                    body = await resp.json()
                except (aiohttp.ContentTypeError, ValueError, TypeError):
                    body = await resp.text()

                try:
                    resp.raise_for_status()
                except aiohttp.ClientResponseError as exc:
                    self._logger.warning(self.dump_response(resp, self._dump_body))
                    if self._raise_on_exc is True:
                        raise  # pragma: no cover

                    return ObjectDict(
                        body=body,
                        status=resp.status,
                        headers={k: v for k, v in resp.headers.items()},
                        exception=exc
                    )

                self._logger.info(self.dump_response(resp, self._dump_body))
                return ObjectDict(
                    body=body,
                    status=resp.status,
                    headers={k: v for k, v in resp.headers.items()}
                )
        except (aiohttp.ClientError, aiohttp.ServerTimeoutError, asyncio.TimeoutError) as exc:
            self._logger.exception(exc)
            if self._raise_on_exc is True:
                raise  # pragma: no cover

            return ObjectDict(
                body={},
                status=httpcode.SERVICE_UNAVAILABLE,
                headers={},
                exception=exc
            )

    async def batch(self, requests):
        """

        :param requests:
        :return:
        """
        tasks = []
        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            for params in requests:
                tasks.append(self.http_request(session, **params))

            return await asyncio.gather(*tasks, return_exceptions=not self._raise_on_exc)

    def request(self, requests, **kwargs):
        """

        :param requests:
        :return:
        """
        if not aiohttp:
            raise ImportError("You must install 'aiohttp'")  # pragma: no cover

        try:
            asyncio.get_running_loop()
            if not nest_asyncio:  # pragma: no cover
                raise ImportError("You must install 'nest-asyncio'")
            # noinspection PyUnresolvedReferences
            nest_asyncio.apply()  # pragma: no cover
        except RuntimeError:
            asyncio.set_event_loop(asyncio.new_event_loop())

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.batch(requests))


class FlaskelHTTPBatch(HTTPBatch, FlaskelHTTPDumper):
    def __init__(self, **kwargs):
        kwargs.setdefault('logger', cap.logger)
        super().__init__(**kwargs)

    def request(self, requests, **kwargs):
        if flask.request.id:
            for r in requests:
                if not r.get('headers'):
                    r['headers'] = {}
                req_id = f"{flask.request.id},{get_uuid()}"
                r['headers'][cap.config.REQUEST_ID_HEADER] = req_id

        return super().request(requests, **kwargs)



from flaskel.utils.datastruct import ObjectDict

SCHEMAS = ObjectDict(
    JSONRPC=ObjectDict(
        REQUEST={
            "$schema":     "http://json-schema.org/draft-04/schema#",
            "description": "A JSON RPC 2.0 request",
            "oneOf":       [
                {
                    "description": "An individual request",
                    "$ref":        "#/definitions/request"
                },
                {
                    "description": "An array of requests",
                    "type":        "array",
                    "items":       {"$ref": "#/definitions/request"}
                }
            ],
            "definitions": {
                "request": {
                    "type":       "object",
                    "required":   ["jsonrpc", "method"],
                    "properties": {
                        "jsonrpc": {"enum": ["2.0"]},
                        "method":  {
                            "type": "string"
                        },
                        "id":      {
                            "type": ["string", "number", "null"],
                            "note": [
                                "While allowed, null should be avoided: http://www.jsonrpc.org/specification#id1",
                                "While allowed, a number with a fractional part should be avoided: "
                                "http://www.jsonrpc.org/specification#id2"
                            ]
                        },
                        "params":  {
                            "type": ["array", "object"]
                        }
                    }
                }
            }
        },
        RESPONSE={
            "$schema":     "http://json-schema.org/draft-07/schema#",
            "oneOf":       [
                {
                    "$ref": "#/definitions/response"
                },
                {
                    "type":  "array",
                    "items": {
                        "$ref": "#/definitions/response"
                    }
                }
            ],
            "definitions": {
                "response": {
                    "type":       [
                        "array",
                        "object"
                    ],
                    "required":   [
                        "jsonrpc"
                    ],
                    "properties": {
                        "jsonrpc": {
                            "enum": [
                                "2.0"
                            ]
                        },
                        "id":      {
                            "type": [
                                "string",
                                "number",
                                "null"
                            ]
                        },
                        "result":  {
                            "type": [
                                "array",
                                "object"
                            ]
                        },
                        "error":   {
                            "type":       [
                                "array",
                                "object"
                            ],
                            "properties": {
                                "code":    {
                                    "type": "number"
                                },
                                "message": {
                                    "type": "string"
                                }
                            }
                        }
                    }
                }
            }
        }
    ),
    HEALTHCHECK={
        "type":       "object",
        "required":   [
            "status",
            "checks",
            "links"
        ],
        "properties": {
            "status": {
                "type": "string"
            },
            "checks": {
                "type":              "object",
                "patternProperties": {
                    ".": {
                        "type":       "object",
                        "required":   [
                            "status",
                            "output"
                        ],
                        "properties": {
                            "status": {
                                "type": "string"
                            },
                            "output": {
                                "type": ["null", "string"]
                            }
                        }
                    }
                }
            },
            "links":  {
                "type":       "object",
                "properties": {
                    "about": {
                        "type": ["null", "string"]
                    }
                }
            }
        }
    },
    API_PROBLEM={
        "$schema":    "http://json-schema.org/draft-07/schema#",
        "type":       "object",
        "required":   [
            "type",
            "title",
            "detail",
            "instance",
            "status"
        ],
        "properties": {
            "type":     {
                "type": "string"
            },
            "title":    {
                "type": "string"
            },
            "detail":   {
                "type": "string"
            },
            "instance": {
                "type": "string"
            },
            "status":   {
                "type": "integer"
            },
            "response": {
                "type": ["null", "object", "array"]
            }
        }
    },
    OPENAPI3={
        "id":                   "https://spec.openapis.org/oas/3.0/schema/2019-04-02",
        "$schema":              "http://json-schema.org/draft-04/schema#",
        "description":          "Validation schema for OpenAPI Specification 3.0.X.",
        "type":                 "object",
        "required":             [
            "openapi",
            "info",
            "paths"
        ],
        "properties":           {
            "openapi":      {
                "type":    "string",
                "pattern": "^3\\.0\\.\\d(-.+)?$"
            },
            "info":         {
                "$ref": "#/definitions/Info"
            },
            "externalDocs": {
                "$ref": "#/definitions/ExternalDocumentation"
            },
            "servers":      {
                "type":  "array",
                "items": {
                    "$ref": "#/definitions/Server"
                }
            },
            "security":     {
                "type":  "array",
                "items": {
                    "$ref": "#/definitions/SecurityRequirement"
                }
            },
            "tags":         {
                "type":        "array",
                "items":       {
                    "$ref": "#/definitions/Tag"
                },
                "uniqueItems": True
            },
            "paths":        {
                "$ref": "#/definitions/Paths"
            },
            "components":   {
                "$ref": "#/definitions/Components"
            }
        },
        "patternProperties":    {
            "^x-": {
            }
        },
        "additionalProperties": False,
        "definitions":          {
            "Reference":                   {
                "type":              "object",
                "required":          [
                    "$ref"
                ],
                "patternProperties": {
                    "^\\$ref$": {
                        "type":   "string",
                        "format": "uri-reference"
                    }
                }
            },
            "Info":                        {
                "type":                 "object",
                "required":             [
                    "title",
                    "version"
                ],
                "properties":           {
                    "title":          {
                        "type": "string"
                    },
                    "description":    {
                        "type": "string"
                    },
                    "termsOfService": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "contact":        {
                        "$ref": "#/definitions/Contact"
                    },
                    "license":        {
                        "$ref": "#/definitions/License"
                    },
                    "version":        {
                        "type": "string"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Contact":                     {
                "type":                 "object",
                "properties":           {
                    "name":  {
                        "type": "string"
                    },
                    "url":   {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "email": {
                        "type":   "string",
                        "format": "email"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "License":                     {
                "type":                 "object",
                "required":             [
                    "name"
                ],
                "properties":           {
                    "name": {
                        "type": "string"
                    },
                    "url":  {
                        "type":   "string",
                        "format": "uri-reference"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Server":                      {
                "type":                 "object",
                "required":             [
                    "url"
                ],
                "properties":           {
                    "url":         {
                        "type": "string"
                    },
                    "description": {
                        "type": "string"
                    },
                    "variables":   {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/ServerVariable"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "ServerVariable":              {
                "type":                 "object",
                "required":             [
                    "default"
                ],
                "properties":           {
                    "enum":        {
                        "type":  "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "default":     {
                        "type": "string"
                    },
                    "description": {
                        "type": "string"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Components":                  {
                "type":                 "object",
                "properties":           {
                    "schemas":         {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Schema"
                                    },
                                    {
                                        "$ref": "#/definitions/Reference"
                                    }
                                ]
                            }
                        }
                    },
                    "responses":       {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/Response"
                                    }
                                ]
                            }
                        }
                    },
                    "parameters":      {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/Parameter"
                                    }
                                ]
                            }
                        }
                    },
                    "examples":        {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/Example"
                                    }
                                ]
                            }
                        }
                    },
                    "requestBodies":   {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/RequestBody"
                                    }
                                ]
                            }
                        }
                    },
                    "headers":         {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/Header"
                                    }
                                ]
                            }
                        }
                    },
                    "securitySchemes": {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/SecurityScheme"
                                    }
                                ]
                            }
                        }
                    },
                    "links":           {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/Link"
                                    }
                                ]
                            }
                        }
                    },
                    "callbacks":       {
                        "type":              "object",
                        "patternProperties": {
                            "^[a-zA-Z0-9\\.\\-_]+$": {
                                "oneOf": [
                                    {
                                        "$ref": "#/definitions/Reference"
                                    },
                                    {
                                        "$ref": "#/definitions/Callback"
                                    }
                                ]
                            }
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Schema":                      {
                "type":                 "object",
                "properties":           {
                    "title":                {
                        "type": "string"
                    },
                    "multipleOf":           {
                        "type":             "number",
                        "minimum":          0,
                        "exclusiveMinimum": True
                    },
                    "maximum":              {
                        "type": "number"
                    },
                    "exclusiveMaximum":     {
                        "type":    "boolean",
                        "default": False
                    },
                    "minimum":              {
                        "type": "number"
                    },
                    "exclusiveMinimum":     {
                        "type":    "boolean",
                        "default": False
                    },
                    "maxLength":            {
                        "type":    "integer",
                        "minimum": 0
                    },
                    "minLength":            {
                        "type":    "integer",
                        "minimum": 0,
                        "default": 0
                    },
                    "pattern":              {
                        "type":   "string",
                        "format": "regex"
                    },
                    "maxItems":             {
                        "type":    "integer",
                        "minimum": 0
                    },
                    "minItems":             {
                        "type":    "integer",
                        "minimum": 0,
                        "default": 0
                    },
                    "uniqueItems":          {
                        "type":    "boolean",
                        "default": False
                    },
                    "maxProperties":        {
                        "type":    "integer",
                        "minimum": 0
                    },
                    "minProperties":        {
                        "type":    "integer",
                        "minimum": 0,
                        "default": 0
                    },
                    "required":             {
                        "type":        "array",
                        "items":       {
                            "type": "string"
                        },
                        "minItems":    1,
                        "uniqueItems": True
                    },
                    "enum":                 {
                        "type":        "array",
                        "items":       {
                        },
                        "minItems":    1,
                        "uniqueItems": False
                    },
                    "type":                 {
                        "type": "string",
                        "enum": [
                            "array",
                            "boolean",
                            "integer",
                            "number",
                            "object",
                            "string"
                        ]
                    },
                    "not":                  {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Schema"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "allOf":                {
                        "type":  "array",
                        "items": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Schema"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "oneOf":                {
                        "type":  "array",
                        "items": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Schema"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "anyOf":                {
                        "type":  "array",
                        "items": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Schema"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "items":                {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Schema"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "properties":           {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Schema"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "additionalProperties": {
                        "oneOf":   [
                            {
                                "$ref": "#/definitions/Schema"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            },
                            {
                                "type": "boolean"
                            }
                        ],
                        "default": True
                    },
                    "description":          {
                        "type": "string"
                    },
                    "format":               {
                        "type": "string"
                    },
                    "default":              {
                    },
                    "nullable":             {
                        "type":    "boolean",
                        "default": False
                    },
                    "discriminator":        {
                        "$ref": "#/definitions/Discriminator"
                    },
                    "readOnly":             {
                        "type":    "boolean",
                        "default": False
                    },
                    "writeOnly":            {
                        "type":    "boolean",
                        "default": False
                    },
                    "example":              {
                    },
                    "externalDocs":         {
                        "$ref": "#/definitions/ExternalDocumentation"
                    },
                    "deprecated":           {
                        "type":    "boolean",
                        "default": False
                    },
                    "xml":                  {
                        "$ref": "#/definitions/XML"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Discriminator":               {
                "type":       "object",
                "required":   [
                    "propertyName"
                ],
                "properties": {
                    "propertyName": {
                        "type": "string"
                    },
                    "mapping":      {
                        "type":                 "object",
                        "additionalProperties": {
                            "type": "string"
                        }
                    }
                }
            },
            "XML":                         {
                "type":                 "object",
                "properties":           {
                    "name":      {
                        "type": "string"
                    },
                    "namespace": {
                        "type":   "string",
                        "format": "uri"
                    },
                    "prefix":    {
                        "type": "string"
                    },
                    "attribute": {
                        "type":    "boolean",
                        "default": False
                    },
                    "wrapped":   {
                        "type":    "boolean",
                        "default": False
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Response":                    {
                "type":                 "object",
                "required":             [
                    "description"
                ],
                "properties":           {
                    "description": {
                        "type": "string"
                    },
                    "headers":     {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Header"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "content":     {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/MediaType"
                        }
                    },
                    "links":       {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Link"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "MediaType":                   {
                "type":                 "object",
                "properties":           {
                    "schema":   {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Schema"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "example":  {
                    },
                    "examples": {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Example"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "encoding": {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/Encoding"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False,
                "allOf":                [
                    {
                        "$ref": "#/definitions/ExampleXORExamples"
                    }
                ]
            },
            "Example":                     {
                "type":                 "object",
                "properties":           {
                    "summary":       {
                        "type": "string"
                    },
                    "description":   {
                        "type": "string"
                    },
                    "value":         {
                    },
                    "externalValue": {
                        "type":   "string",
                        "format": "uri-reference"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Header":                      {
                "type":                 "object",
                "properties":           {
                    "description":     {
                        "type": "string"
                    },
                    "required":        {
                        "type":    "boolean",
                        "default": False
                    },
                    "deprecated":      {
                        "type":    "boolean",
                        "default": False
                    },
                    "allowEmptyValue": {
                        "type":    "boolean",
                        "default": False
                    },
                    "style":           {
                        "type":    "string",
                        "enum":    [
                            "simple"
                        ],
                        "default": "simple"
                    },
                    "explode":         {
                        "type": "boolean"
                    },
                    "allowReserved":   {
                        "type":    "boolean",
                        "default": False
                    },
                    "schema":          {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Schema"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "content":         {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/MediaType"
                        },
                        "minProperties":        1,
                        "maxProperties":        1
                    },
                    "example":         {
                    },
                    "examples":        {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Example"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False,
                "allOf":                [
                    {
                        "$ref": "#/definitions/ExampleXORExamples"
                    },
                    {
                        "$ref": "#/definitions/SchemaXORContent"
                    }
                ]
            },
            "Paths":                       {
                "type":                 "object",
                "patternProperties":    {
                    "^\\/": {
                        "$ref": "#/definitions/PathItem"
                    },
                    "^x-":  {
                    }
                },
                "additionalProperties": False
            },
            "PathItem":                    {
                "type":                 "object",
                "properties":           {
                    "$ref":        {
                        "type": "string"
                    },
                    "summary":     {
                        "type": "string"
                    },
                    "description": {
                        "type": "string"
                    },
                    "servers":     {
                        "type":  "array",
                        "items": {
                            "$ref": "#/definitions/Server"
                        }
                    },
                    "parameters":  {
                        "type":        "array",
                        "items":       {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Parameter"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        },
                        "uniqueItems": True
                    }
                },
                "patternProperties":    {
                    "^(get|put|post|delete|options|head|patch|trace)$": {
                        "$ref": "#/definitions/Operation"
                    },
                    "^x-":                                              {
                    }
                },
                "additionalProperties": False
            },
            "Operation":                   {
                "type":                 "object",
                "required":             [
                    "responses"
                ],
                "properties":           {
                    "tags":         {
                        "type":  "array",
                        "items": {
                            "type": "string"
                        }
                    },
                    "summary":      {
                        "type": "string"
                    },
                    "description":  {
                        "type": "string"
                    },
                    "externalDocs": {
                        "$ref": "#/definitions/ExternalDocumentation"
                    },
                    "operationId":  {
                        "type": "string"
                    },
                    "parameters":   {
                        "type":        "array",
                        "items":       {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Parameter"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        },
                        "uniqueItems": True
                    },
                    "requestBody":  {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/RequestBody"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "responses":    {
                        "$ref": "#/definitions/Responses"
                    },
                    "callbacks":    {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Callback"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    },
                    "deprecated":   {
                        "type":    "boolean",
                        "default": False
                    },
                    "security":     {
                        "type":  "array",
                        "items": {
                            "$ref": "#/definitions/SecurityRequirement"
                        }
                    },
                    "servers":      {
                        "type":  "array",
                        "items": {
                            "$ref": "#/definitions/Server"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Responses":                   {
                "type":                 "object",
                "properties":           {
                    "default": {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Response"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    }
                },
                "patternProperties":    {
                    "^[1-5](?:\\d{2}|XX)$": {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Response"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "^x-":                  {
                    }
                },
                "minProperties":        1,
                "additionalProperties": False
            },
            "SecurityRequirement":         {
                "type":                 "object",
                "additionalProperties": {
                    "type":  "array",
                    "items": {
                        "type": "string"
                    }
                }
            },
            "Tag":                         {
                "type":                 "object",
                "required":             [
                    "name"
                ],
                "properties":           {
                    "name":         {
                        "type": "string"
                    },
                    "description":  {
                        "type": "string"
                    },
                    "externalDocs": {
                        "$ref": "#/definitions/ExternalDocumentation"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "ExternalDocumentation":       {
                "type":                 "object",
                "required":             [
                    "url"
                ],
                "properties":           {
                    "description": {
                        "type": "string"
                    },
                    "url":         {
                        "type":   "string",
                        "format": "uri-reference"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "ExampleXORExamples":          {
                "description": "Example and examples are mutually exclusive",
                "not":         {
                    "required": [
                        "example",
                        "examples"
                    ]
                }
            },
            "SchemaXORContent":            {
                "description": "Schema and content are mutually exclusive, at least one is required",
                "not":         {
                    "required": [
                        "schema",
                        "content"
                    ]
                },
                "oneOf":       [
                    {
                        "required": [
                            "schema"
                        ]
                    },
                    {
                        "required":    [
                            "content"
                        ],
                        "description": "Some properties are not allowed if content is present",
                        "allOf":       [
                            {
                                "not": {
                                    "required": [
                                        "style"
                                    ]
                                }
                            },
                            {
                                "not": {
                                    "required": [
                                        "explode"
                                    ]
                                }
                            },
                            {
                                "not": {
                                    "required": [
                                        "allowReserved"
                                    ]
                                }
                            },
                            {
                                "not": {
                                    "required": [
                                        "example"
                                    ]
                                }
                            },
                            {
                                "not": {
                                    "required": [
                                        "examples"
                                    ]
                                }
                            }
                        ]
                    }
                ]
            },
            "Parameter":                   {
                "type":                 "object",
                "properties":           {
                    "name":            {
                        "type": "string"
                    },
                    "in":              {
                        "type": "string"
                    },
                    "description":     {
                        "type": "string"
                    },
                    "required":        {
                        "type":    "boolean",
                        "default": False
                    },
                    "deprecated":      {
                        "type":    "boolean",
                        "default": False
                    },
                    "allowEmptyValue": {
                        "type":    "boolean",
                        "default": False
                    },
                    "style":           {
                        "type": "string"
                    },
                    "explode":         {
                        "type": "boolean"
                    },
                    "allowReserved":   {
                        "type":    "boolean",
                        "default": False
                    },
                    "schema":          {
                        "oneOf": [
                            {
                                "$ref": "#/definitions/Schema"
                            },
                            {
                                "$ref": "#/definitions/Reference"
                            }
                        ]
                    },
                    "content":         {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/MediaType"
                        },
                        "minProperties":        1,
                        "maxProperties":        1
                    },
                    "example":         {
                    },
                    "examples":        {
                        "type":                 "object",
                        "additionalProperties": {
                            "oneOf": [
                                {
                                    "$ref": "#/definitions/Example"
                                },
                                {
                                    "$ref": "#/definitions/Reference"
                                }
                            ]
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False,
                "required":             [
                    "name",
                    "in"
                ],
                "allOf":                [
                    {
                        "$ref": "#/definitions/ExampleXORExamples"
                    },
                    {
                        "$ref": "#/definitions/SchemaXORContent"
                    },
                    {
                        "$ref": "#/definitions/ParameterLocation"
                    }
                ]
            },
            "ParameterLocation":           {
                "description": "Parameter location",
                "oneOf":       [
                    {
                        "description": "Parameter in path",
                        "required":    [
                            "required"
                        ],
                        "properties":  {
                            "in":       {
                                "enum": [
                                    "path"
                                ]
                            },
                            "style":    {
                                "enum":    [
                                    "matrix",
                                    "label",
                                    "simple"
                                ],
                                "default": "simple"
                            },
                            "required": {
                                "enum": [
                                    True
                                ]
                            }
                        }
                    },
                    {
                        "description": "Parameter in query",
                        "properties":  {
                            "in":    {
                                "enum": [
                                    "query"
                                ]
                            },
                            "style": {
                                "enum":    [
                                    "form",
                                    "spaceDelimited",
                                    "pipeDelimited",
                                    "deepObject"
                                ],
                                "default": "form"
                            }
                        }
                    },
                    {
                        "description": "Parameter in header",
                        "properties":  {
                            "in":    {
                                "enum": [
                                    "header"
                                ]
                            },
                            "style": {
                                "enum":    [
                                    "simple"
                                ],
                                "default": "simple"
                            }
                        }
                    },
                    {
                        "description": "Parameter in cookie",
                        "properties":  {
                            "in":    {
                                "enum": [
                                    "cookie"
                                ]
                            },
                            "style": {
                                "enum":    [
                                    "form"
                                ],
                                "default": "form"
                            }
                        }
                    }
                ]
            },
            "RequestBody":                 {
                "type":                 "object",
                "required":             [
                    "content"
                ],
                "properties":           {
                    "description": {
                        "type": "string"
                    },
                    "content":     {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/MediaType"
                        }
                    },
                    "required":    {
                        "type":    "boolean",
                        "default": False
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "SecurityScheme":              {
                "oneOf": [
                    {
                        "$ref": "#/definitions/APIKeySecurityScheme"
                    },
                    {
                        "$ref": "#/definitions/HTTPSecurityScheme"
                    },
                    {
                        "$ref": "#/definitions/OAuth2SecurityScheme"
                    },
                    {
                        "$ref": "#/definitions/OpenIdConnectSecurityScheme"
                    }
                ]
            },
            "APIKeySecurityScheme":        {
                "type":                 "object",
                "required":             [
                    "type",
                    "name",
                    "in"
                ],
                "properties":           {
                    "type":        {
                        "type": "string",
                        "enum": [
                            "apiKey"
                        ]
                    },
                    "name":        {
                        "type": "string"
                    },
                    "in":          {
                        "type": "string",
                        "enum": [
                            "header",
                            "query",
                            "cookie"
                        ]
                    },
                    "description": {
                        "type": "string"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "HTTPSecurityScheme":          {
                "type":                 "object",
                "required":             [
                    "scheme",
                    "type"
                ],
                "properties":           {
                    "scheme":       {
                        "type": "string"
                    },
                    "bearerFormat": {
                        "type": "string"
                    },
                    "description":  {
                        "type": "string"
                    },
                    "type":         {
                        "type": "string",
                        "enum": [
                            "http"
                        ]
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False,
                "oneOf":                [
                    {
                        "description": "Bearer",
                        "properties":  {
                            "scheme": {
                                "enum": [
                                    "bearer"
                                ]
                            }
                        }
                    },
                    {
                        "description": "Non Bearer",
                        "not":         {
                            "required": [
                                "bearerFormat"
                            ]
                        },
                        "properties":  {
                            "scheme": {
                                "not": {
                                    "enum": [
                                        "bearer"
                                    ]
                                }
                            }
                        }
                    }
                ]
            },
            "OAuth2SecurityScheme":        {
                "type":                 "object",
                "required":             [
                    "type",
                    "flows"
                ],
                "properties":           {
                    "type":        {
                        "type": "string",
                        "enum": [
                            "oauth2"
                        ]
                    },
                    "flows":       {
                        "$ref": "#/definitions/OAuthFlows"
                    },
                    "description": {
                        "type": "string"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "OpenIdConnectSecurityScheme": {
                "type":                 "object",
                "required":             [
                    "type",
                    "openIdConnectUrl"
                ],
                "properties":           {
                    "type":             {
                        "type": "string",
                        "enum": [
                            "openIdConnect"
                        ]
                    },
                    "openIdConnectUrl": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "description":      {
                        "type": "string"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "OAuthFlows":                  {
                "type":                 "object",
                "properties":           {
                    "implicit":          {
                        "$ref": "#/definitions/ImplicitOAuthFlow"
                    },
                    "password":          {
                        "$ref": "#/definitions/PasswordOAuthFlow"
                    },
                    "clientCredentials": {
                        "$ref": "#/definitions/ClientCredentialsFlow"
                    },
                    "authorizationCode": {
                        "$ref": "#/definitions/AuthorizationCodeOAuthFlow"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "ImplicitOAuthFlow":           {
                "type":                 "object",
                "required":             [
                    "authorizationUrl",
                    "scopes"
                ],
                "properties":           {
                    "authorizationUrl": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "refreshUrl":       {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "scopes":           {
                        "type":                 "object",
                        "additionalProperties": {
                            "type": "string"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "PasswordOAuthFlow":           {
                "type":                 "object",
                "required":             [
                    "tokenUrl"
                ],
                "properties":           {
                    "tokenUrl":   {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "refreshUrl": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "scopes":     {
                        "type":                 "object",
                        "additionalProperties": {
                            "type": "string"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "ClientCredentialsFlow":       {
                "type":                 "object",
                "required":             [
                    "tokenUrl"
                ],
                "properties":           {
                    "tokenUrl":   {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "refreshUrl": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "scopes":     {
                        "type":                 "object",
                        "additionalProperties": {
                            "type": "string"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "AuthorizationCodeOAuthFlow":  {
                "type":                 "object",
                "required":             [
                    "authorizationUrl",
                    "tokenUrl"
                ],
                "properties":           {
                    "authorizationUrl": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "tokenUrl":         {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "refreshUrl":       {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "scopes":           {
                        "type":                 "object",
                        "additionalProperties": {
                            "type": "string"
                        }
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False
            },
            "Link":                        {
                "type":                 "object",
                "properties":           {
                    "operationId":  {
                        "type": "string"
                    },
                    "operationRef": {
                        "type":   "string",
                        "format": "uri-reference"
                    },
                    "parameters":   {
                        "type":                 "object",
                        "additionalProperties": {
                        }
                    },
                    "requestBody":  {
                    },
                    "description":  {
                        "type": "string"
                    },
                    "server":       {
                        "$ref": "#/definitions/Server"
                    }
                },
                "patternProperties":    {
                    "^x-": {
                    }
                },
                "additionalProperties": False,
                "not":                  {
                    "description": "Operation Id and Operation Ref are mutually exclusive",
                    "required":    [
                        "operationId",
                        "operationRef"
                    ]
                }
            },
            "Callback":                    {
                "type":                 "object",
                "additionalProperties": {
                    "$ref": "#/definitions/PathItem"
                },
                "patternProperties":    {
                    "^x-": {
                    }
                }
            },
            "Encoding":                    {
                "type":                 "object",
                "properties":           {
                    "contentType":   {
                        "type": "string"
                    },
                    "headers":       {
                        "type":                 "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/Header"
                        }
                    },
                    "style":         {
                        "type": "string",
                        "enum": [
                            "form",
                            "spaceDelimited",
                            "pipeDelimited",
                            "deepObject"
                        ]
                    },
                    "explode":       {
                        "type": "boolean"
                    },
                    "allowReserved": {
                        "type":    "boolean",
                        "default": False
                    }
                },
                "additionalProperties": False
            }
        }
    }
)



from .base import WSGIBuiltin

try:
    from .gevent import WSGIGevent
except ImportError:
    WSGIGevent = None
try:
    from .gunicorn import WSGIGunicorn
except ImportError:
    WSGIGunicorn = None
try:
    from .tornado import WSGITornado
except ImportError:
    WSGITornado = None
try:
    from .twisted import WSGITwisted
except ImportError:
    WSGITwisted = None
try:
    from .waitress import WSGIWaitress
except ImportError:
    WSGIWaitress = None


class WSGIFactory:
    WSGI_SERVERS = {
        'builtin':  WSGIBuiltin,
        'gunicorn': WSGIGunicorn,
        'gevent':   WSGIGevent,
        'tornado':  WSGITornado,
        'twisted':  WSGITwisted,
        'waitress': WSGIWaitress,
    }

    @classmethod
    def get_instance(cls, name, **kwargs):
        """

        :param name:
        :param kwargs:
        :return:
        """
        wsgi_class = cls.get_class(name)
        return wsgi_class(**kwargs)

    @classmethod
    def get_class(cls, name):
        """

        :param name:
        :return:
        """
        if name not in cls.WSGI_SERVERS:
            raise ValueError("unable to find wsgi server: '{}'".format(name))

        wsgi_class = cls.WSGI_SERVERS.get(name)

        if wsgi_class is None:
            raise ImportError("You must install '{}' dependencies".format(name))

        return wsgi_class



from celery import Celery


class FlaskCelery:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        :param kwargs:
        """
        self.session = Celery(**kwargs)

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """

        :param app:
        :return:
        """
        if not hasattr(app, 'extensions'):
            app.extensions = {}  # pragma: no cover
        app.extensions['celery'] = self

        self.session.conf.update(app.config.get('CELERY') or {})
        app.logger.debug(f"Celery configuration\n:{dict(self.session.conf)}")

        class ContextTask(self.session.Task):
            abstract = True

            def __call__(self, *args, **kw):
                with app.app_context():
                    return self.celery.Task.__call__(self, *args, **kw)

        # noinspection PyPropertyAccess
        self.session.Task = ContextTask

    @staticmethod
    def celery_app(app=None, factory=None, conf=None):
        """

        :param app:
        :param factory:
        :param conf:
        :return:
        """
        _app = app or factory.get_or_create(conf)
        _celery = _app.extensions['celery']
        return _celery.session


celery = FlaskCelery()



import flask

from flaskel import cap, httpcode
from flaskel.utils import uuid
from . import test


@test.route('/method_override', methods=['POST', 'PUT'])
def method_override_post():
    return '', httpcode.SUCCESS if flask.request.method == 'PUT' else httpcode.METHOD_NOT_ALLOWED


@test.route('/test_https')
def test_https():
    remote = cap.extensions['cloudflareRemote']
    return {
        'address': remote.get_client_ip(),
        'scheme':  flask.request.scheme,
        'url_for': flask.url_for('test.test_https', _external=True)
    }


@test.route('/proxy')
def test_proxy():
    return {
        'script_name': flask.request.environ['SCRIPT_NAME'],
        'original':    flask.request.environ['werkzeug.proxy_fix.orig']
    }


@test.route("/list/<list('-'):data>")
def list_converter(data):
    return flask.jsonify(data)


@test.route('/invalid-json', methods=['POST'])
def get_invalid_json():
    flask.request.get_json()
    return '', httpcode.SUCCESS


@test.route('/download')
def download():
    response = cap.response_class
    return response.send_file('./', flask.request.args.get('filename'))


@test.route('/uuid')
def return_uuid():
    return flask.jsonify(dict(
        uuid1=uuid.get_uuid(ver=1),
        uuid3=uuid.get_uuid(ver=3),
        uuid4=uuid.get_uuid(),
        uuid5=uuid.get_uuid(ver=5),
    ))


@test.route('/crypt/<passwd>')
def crypt(passwd):
    crypto = cap.extensions['argon2']
    return crypto.generate_hash(passwd)


@test.route('/useragent')
def useragent():
    return flask.g.user_agent.to_dict()



import flask
from flask.views import MethodView, View

from flaskel import httpcode
from flaskel.ext import builder


class BaseView(View):
    methods = ['GET']

    def dispatch_request(self):
        """
        Must be implemented in every subclass
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def register(cls, app, name, *urls, **kwargs):
        """

        :param app: Flask app or blueprint
        :param name: optional view name
        :param urls: optional urls or view name
        :param kwargs: argument passed to cls constructor
        """
        name = name or cls.__name__
        view_func = cls.as_view(name, **kwargs)
        for u in urls or (f"/{name}",):
            app.add_url_rule(u, view_func=view_func, methods=cls.methods)


class Resource(MethodView):
    @builder.on_accept()
    def get(self, *args, res_id=None, sub_resource=None, **kwargs):
        """

        :param res_id:
        :param sub_resource:
        :return:
        """
        if res_id is None:
            return self.on_collection(*args, **kwargs)

        if sub_resource is None:
            return self.on_get(res_id, *args, **kwargs)

        try:
            _sub_resource = getattr(self, "sub_{}".format(sub_resource))
            return _sub_resource(res_id, *args, **kwargs)
        except AttributeError:
            flask.abort(httpcode.NOT_FOUND)

    @builder.on_accept()
    def post(self, *args, res_id=None, sub_resource=None, **kwargs):
        """

        :param res_id:
        :param sub_resource
        :return:
        """
        if res_id is None:
            return self.on_post(*args, **kwargs)
        try:
            _sub_resource_post = getattr(self, "sub_{}_post".format(sub_resource))
            return _sub_resource_post(res_id, *args, **kwargs)
        except AttributeError:
            flask.abort(httpcode.NOT_FOUND)

    @builder.no_content
    def delete(self, res_id, *args, **kwargs):
        """

        :param res_id:
        :return:
        """
        return self.on_delete(res_id, *args, **kwargs)

    @builder.on_accept()
    def put(self, res_id, *args, **kwargs):
        """

        :param res_id:
        :return:
        """
        return self.on_put(res_id, *args, **kwargs)

    def on_get(self, res_id, *args, **kwargs):
        self._not_implemented()  # pragma: no cover

    def on_post(self, *args, **kwargs):
        self._not_implemented()  # pragma: no cover

    def on_put(self, res_id, *args, **kwargs):
        self._not_implemented()  # pragma: no cover

    def on_delete(self, res_id, *args, **kwargs):
        self._not_implemented()  # pragma: no cover

    def on_collection(self, *args, **kwargs):
        self._not_implemented()  # pragma: no cover

    # noinspection PyMethodMayBeStatic
    def _not_implemented(self):  # pragma: no cover
        flask.abort(httpcode.NOT_IMPLEMENTED)

    @classmethod
    def register(cls, app, name=None, url=None, pk_type='int', **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name: view name
        :param url: url to bind
        :param pk_type: type of res_id
        """
        name = name or cls.__name__
        url = url or f"/{name}"
        view_func = cls.as_view(name, **kwargs)
        url = f"/{url.rstrip('/')}"

        app.add_url_rule(
            url,
            view_func=view_func,
            methods=['GET', 'POST']
        )
        app.add_url_rule(
            f"{url}/<{pk_type}:res_id>",
            view_func=view_func,
            methods=['GET', 'PUT', 'DELETE']
        )
        app.add_url_rule(
            f"{url}/<{pk_type}:res_id>/<sub_resource>",
            view_func=view_func,
            methods=['GET', 'POST']
        )



import os

import pytest

from blueprints import BLUEPRINTS
from ext import EXTENSIONS
from flaskel import TestClient
from flaskel.ext import BASE_EXTENSIONS
from flaskel.utils import yaml

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
CONF_FILE = os.path.join(BASE_DIR, 'config', 'conf.json')

EXTRAS = dict(
    template_folder="blueprints/web/templates",
    static_folder="blueprints/web/static"
)


@pytest.fixture(scope='session')
def testapp():
    return TestClient.get_app(
        conf=yaml.load_yaml_file(CONF_FILE),
        blueprints=BLUEPRINTS,
        extensions={**BASE_EXTENSIONS, **EXTENSIONS},
        **(EXTRAS or {})
    ).test_client()



import time

from flaskel.http import HTTPClient


class FetchMail(HTTPClient):
    """
    Fetch emails from sendria
    """

    def __init__(self, endpoint, auth=None, retry=3, wait=1, timeout=3):
        """

        :param endpoint:
        :param auth:
        :param retry:
        :param wait:
        :param timeout:
        """
        super().__init__(endpoint, auth, timeout)
        self.retry = retry
        self.wait = wait

    def get(self, recipient, subject=None):
        """

        :param recipient:
        :param subject:
        :return:
        """
        resp = self.request('/')

        response = []
        for d in resp.data:
            if recipient in d.recipients_envelope:
                if subject and d.subject != subject:
                    continue
                response.append(d)

        return response

    def perform(self, recipient=None, subject=None, delete=True):
        """

        :param recipient:
        :param subject:
        :param delete:
        :return:
        """
        for _ in range(0, self.retry):
            res = self.get(recipient, subject)
            if delete:
                for r in res:
                    self.request(uri=f"/{r.id}", method="DELETE")
            if res:
                return res

            time.sleep(self.wait)



import os
from base64 import b64encode
from functools import partial

import flask
import werkzeug.exceptions

from flaskel import datastruct, datetime, http, httpcode, SCHEMAS, uuid, yaml
from flaskel.http import batch, rpc, useragent
from flaskel.tester import Asserter
# noinspection PyUnresolvedReferences
from . import app_dev, app_prod, testapp

yaml.setup_yaml_parser()

url_for = partial(flask.url_for, _external=True)

CTS = datastruct.ObjectDict(
    json='application/json',
    xml='application/xml',
    html='text/html',
    json_problem='application/problem+json',
    xml_problem='application/problem+xml',
    json_health='application/health+json',
)

HOSTS = datastruct.ObjectDict(
    apitester="http://httpbin.org",
    fake="http://localhost"
)


def test_app_dev(app_dev):
    client = app_dev.test_client()
    res = client.get(flask.url_for('web.index'))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.html)
    Asserter.assert_equals(app_dev.config.FLASK_ENV, 'development')
    Asserter.assert_equals(app_dev.config.SECRET_KEY, 'fake_very_complex_string')

    res = client.get(url_for('test.test_https'))
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.scheme, 'http')
    Asserter.assert_true(res.json.url_for.startswith('http'))


def test_api_resources(testapp):
    res = testapp.get(url_for('api.resource_api'), headers={'Accept': CTS.xml})
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.xml)

    res = testapp.get(url_for('api.resource_api', res_id=1))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.json)

    res = testapp.get(url_for('api.resource_api', res_id=1, sub_resource='items'))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.json)

    res = testapp.get(url_for('api.resource_api', res_id=1, sub_resource='not-found'))
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.json_problem)
    Asserter.assert_schema(res.json, SCHEMAS.API_PROBLEM)

    res = testapp.delete(url_for('api.resource_api', res_id=1))
    Asserter.assert_status_code(res, httpcode.NO_CONTENT)

    data = dict(item='test')
    res = testapp.put(url_for('api.resource_api', res_id=1), json=data)
    Asserter.assert_status_code(res)

    res = testapp.post(url_for('api.resource_api'), json=data)
    Asserter.assert_status_code(res, httpcode.CREATED)

    res = testapp.post(url_for('api.resource_api', res_id=1, sub_resource='items'), json=data)
    Asserter.assert_status_code(res, httpcode.CREATED)


def test_api_cors(testapp):
    res = testapp.get(url_for('api.resource_api'))
    Asserter.assert_header(res, 'Access-Control-Allow-Origin', '*')


def test_dispatch_error_web(testapp):
    res = testapp.get('/web-page-not-found')
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.html)


def test_dispatch_error_api(testapp):
    res = testapp.get(f"http://api.{testapp.application.config.SERVER_NAME}/not_found")
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.json_problem)
    Asserter.assert_schema(res.json, SCHEMAS.API_PROBLEM)


def test_force_https(testapp):
    res = testapp.get(url_for('test.test_https'))
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.scheme, 'https')
    Asserter.assert_true(res.json.url_for.startswith('https'))


def test_reverse_proxy(testapp):
    res = testapp.get(url_for('test.test_proxy'), headers={
        'X-Forwarded-Prefix': '/test'
    })
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.script_name, '/test')
    Asserter.assert_equals(res.json.original.SCRIPT_NAME, '')


def test_secret_key_prod(testapp):
    Asserter.assert_equals(testapp.application.config.FLASK_ENV, 'production')
    Asserter.assert_true(os.path.isfile('.secret.key'))
    os.remove('.secret.key')


def test_method_override(testapp):
    res = testapp.post(
        url_for('test.method_override_post'),
        headers={'X-HTTP-Method-Override': 'PUT'}
    )
    Asserter.assert_status_code(res)

    res = testapp.post(f"{url_for('test.method_override_post')}?_method_override=PUT")
    Asserter.assert_status_code(res)


def test_converters(testapp):
    res = testapp.get(url_for('test.list_converter', data=['a', 'b', 'c']))
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json, 3)


def test_utils_get_json(testapp):
    res = testapp.post(url_for('test.get_invalid_json'))
    Asserter.assert_status_code(res, httpcode.BAD_REQUEST)


def test_utils_send_file(testapp):
    filename = 'MANIFEST.in'
    url = url_for('test.download') + "?filename={}"

    res = testapp.get(url.format(filename))
    Asserter.assert_status_code(res)
    Asserter.assert_header(res, 'Content-Disposition', f'attachment; filename={filename}')
    Asserter.assert_header(res, 'X-Accel-Redirect', f'./{filename}')
    Asserter.assert_true(res.headers.get('X-Sendfile').endswith(f'./{filename}'))
    Asserter.assert_equals(res.data, b'')

    res = testapp.get(url.format('nofile.txt'))
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.html)


def test_utils_uuid(testapp):
    res = testapp.get(url_for('test.return_uuid'))
    Asserter.assert_false(uuid.check_uuid('fake uuid'))
    Asserter.assert_true(uuid.check_uuid(uuid.get_uuid(hexify=False)))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid1'), ver=1))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid3'), ver=3))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid4')))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid5'), ver=5))


def test_crypto(testapp):
    passwd = 'my-favourite-password'
    crypto = testapp.application.extensions['argon2']
    res = testapp.get(url_for('test.crypt', passwd=passwd))
    Asserter.assert_true(crypto.verify_hash(res.data, passwd))
    Asserter.assert_false(crypto.verify_hash(res.data, "wrong-pass"))


def test_utils_http_client_simple(testapp):
    with testapp.application.test_request_context():
        api = http.FlaskelHttp(HOSTS.apitester, token='pippo', dump_body=True)
        res = api.delete('/status/202')
        Asserter.assert_equals(res.status, httpcode.ACCEPTED)
        res = api.patch('/status/400')
        Asserter.assert_equals(res.status, httpcode.BAD_REQUEST)


def test_utils_http_client_exception(testapp):
    logger = testapp.application.logger
    api = http.HTTPClient(HOSTS.apitester, token='pippo', raise_on_exc=True, logger=logger)
    fake_api = http.HTTPClient(HOSTS.fake, username='test', password='test', logger=logger)

    res = fake_api.put('/', timeout=0.1)
    Asserter.assert_equals(res.status, httpcode.SERVICE_UNAVAILABLE)
    try:
        api.request('/status/500', 'PUT')
    except http.client.http_exc.HTTPError as exc:
        Asserter.assert_equals(exc.response.status_code, httpcode.INTERNAL_SERVER_ERROR)
    try:
        fake_api.request('/', timeout=0.1)
    except werkzeug.exceptions.HTTPException as exc:
        Asserter.assert_equals(exc.code, httpcode.INTERNAL_SERVER_ERROR)


def test_utils_http_client_filename(testapp):
    filename = "pippo.txt"
    hdr = 'Content-Disposition'
    param = {hdr: None}

    api = http.HTTPClient(HOSTS.apitester, dump_body=True, logger=testapp.application.logger)

    param[hdr] = f"attachment; filename={filename}"
    res = api.get('/response-headers', params=param)
    Asserter.assert_equals(api.response_filename(res.headers), filename)

    param[hdr] = f"filename={filename}"
    res = api.post('/response-headers', params=param)
    Asserter.assert_equals(api.response_filename(res.headers), filename)

    param[hdr] = filename
    res = api.post('/response-headers', params=param)
    Asserter.assert_none(api.response_filename(res.headers))


def test_http_client_batch(testapp):
    with testapp.application.test_request_context():
        responses = batch.FlaskelHTTPBatch(dump_body=True).request([
            dict(url=f"{HOSTS.apitester}/anything", method="GET", headers={"HDR1": "HDR1"}),
            dict(url=f"{HOSTS.apitester}/status/{httpcode.NOT_FOUND}", method="GET"),
            dict(url=HOSTS.fake, method='GET', timeout=0.1),
        ])
    Asserter.assert_equals(responses[0].body.headers.Hdr1, 'HDR1')
    Asserter.assert_equals(responses[1].status, httpcode.NOT_FOUND)
    Asserter.assert_equals(responses[2].status, httpcode.SERVICE_UNAVAILABLE)


def test_utils_http_jsonrpc_client(testapp):
    params = dict(a=1, b=2)
    api = http.JsonRPCClient(HOSTS.apitester, "/anything", logger=testapp.application.logger)
    res = api.request('method.test', params=params)
    Asserter.assert_equals(res.json.jsonrpc, '2.0')
    Asserter.assert_equals(res.json.id, api.request_id)
    Asserter.assert_equals(res.json.params, params)


def test_healthcheck(testapp):
    res = testapp.get(testapp.application.config.HEALTHCHECK_PATH)
    Asserter.assert_status_code(res, httpcode.SERVICE_UNAVAILABLE)
    Asserter.assert_content_type(res, CTS.json_health)
    Asserter.assert_schema(res.json, SCHEMAS.HEALTHCHECK)


def test_api_jsonrpc_success(testapp):
    call_id = 1
    url = url_for('api.myJsonRPC')

    res = testapp.jsonrpc(url, method="MyJsonRPC.testAction1", call_id=call_id)
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.id, call_id)
    Asserter.assert_true(res.json.result.action1)

    res = testapp.jsonrpc(url, method="MyJsonRPC.testAction2")
    Asserter.assert_status_code(res, httpcode.NO_CONTENT)


def test_api_jsonrpc_error(testapp):
    call_id = 1
    url = url_for('api.myJsonRPC')

    res = testapp.jsonrpc(url, method="NotFoundMethod", call_id=call_id)
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.error.code, rpc.RPCMethodNotFound().code)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.id, call_id)
    Asserter.assert_true(res.json.error.message)

    res = testapp.jsonrpc(url, json={})
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.error.code, rpc.RPCParseError().code)

    res = testapp.jsonrpc(url, json={"params": None})
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInvalidRequest().code)

    res = testapp.jsonrpc(url, json={"jsonrpc": 1, 'method': "MyJsonRPC.testAction1"})
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInvalidRequest().code)

    res = testapp.jsonrpc(url, method="MyJsonRPC.testInternalError", call_id=call_id)
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInternalError().code)


def test_api_jsonrpc_params(testapp):
    url = url_for('api.myJsonRPC')
    method = "MyJsonRPC.testInvalidParams"

    res = testapp.jsonrpc(url, method=method, call_id=1, params={"param": "testparam"})
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_not_in('error', res.json)

    res = testapp.jsonrpc(url, method=method, call_id=1, params={"params": None})
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, SCHEMAS.JSONRPC_RESPONSE)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInvalidParams().code)


def test_useragent(testapp):
    res = testapp.get(url_for('test.useragent'))
    Asserter.assert_status_code(res)
    Asserter.assert_allin(res.json.keys(), ('browser', 'device', 'os', 'raw'))
    ua_string = """
        Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) 
        Chrome/70.0.3538.77 Safari/537.36
    """
    res = useragent.UserAgent().parse(ua_string)
    Asserter.assert_equals(res.raw, ua_string)


def test_utils_date_conversion():
    exc = False
    fmt = "%d %B %Y %I:%M %p"
    iso_date = '2020-12-28T19:53:00'
    not_iso_date = '28 December 2020 07:53 PM'
    invalid_date = 'invalid_date'

    res = datetime.from_iso_format(iso_date, fmt, exc=exc)
    Asserter.assert_true(res)
    Asserter.assert_equals(not_iso_date, res)

    res = datetime.to_iso_format(not_iso_date, fmt, exc=exc)
    Asserter.assert_true(res)
    Asserter.assert_equals(iso_date, res)

    Asserter.assert_none(datetime.from_iso_format(invalid_date, fmt, exc=exc))
    Asserter.assert_none(datetime.to_iso_format(invalid_date, fmt, exc=exc))


def test_correlation_id(testapp):
    cap = testapp.application
    request_id_header = cap.config.REQUEST_ID_HEADER

    res = testapp.get('/')
    Asserter.assert_header(res, request_id_header, is_in=True)
    Asserter.assert_type(res.headers[request_id_header], str)

    req_id = uuid.get_uuid()
    res = testapp.get('/', headers={request_id_header: req_id})
    Asserter.assert_header(res, request_id_header, req_id)

    req_id = 'invalid_req_id'
    res = testapp.get('/', headers={request_id_header: req_id})
    Asserter.assert_different(res.headers[request_id_header], req_id)


def test_jwt(testapp):
    user = testapp.application.config.BASIC_AUTH_USERNAME
    passwd = testapp.application.config.BASIC_AUTH_PASSWORD
    credential = b64encode(f"{user}:{passwd}".encode()).decode()

    res = testapp.post(
        url_for('auth.access_token'),
        json=dict(email='email', password='password'),
        headers={"Authorization": f"Basic {credential}"}
    )
    Asserter.assert_status_code(res)
    Asserter.assert_allin(res.json.keys(), (
        'access_token', 'refresh_token', 'expires_in', 'issued_at', 'token_type', 'scope'
    ))

    res = testapp.post(
        url_for('auth.refresh_token'),
        headers={'Authorization': f"Bearer {res.json.refresh_token}"}
    )
    Asserter.assert_status_code(res)
    Asserter.assert_allin(res.json.keys(), (
        'access_token', 'expires_in', 'issued_at', 'token_type', 'scope'
    ))



import flask
from flask import current_app as cap

from . import auth
from flaskel.ext.auth import jwt, create_tokens, refresh_access_token
from flaskel import httpcode


@auth.route('/access_token', methods=['POST'])
def access_token():
    """

    :return:
    """
    data = flask.request.json
    if data.email == 'email' and data.password == 'password':
        return create_tokens(data.email)

    flask.abort(httpcode.UNAUTHORIZED)


@auth.route('/refresh_token', methods=['GET'])
@jwt.jwt_refresh_token_required
def refresh_token():
    return refresh_access_token()


@auth.route('/check_token', methods=['GET'])
@jwt.jwt_required
def token_check():
    return dict(
        token_type=cap.config['JWT_DEFAULT_TOKEN_TYPE'],
        scope=cap.config['JWT_DEFAULT_SCOPE'],
        **jwt.get_raw_jwt()
    )



from flask_sqlalchemy import Model


class SQLAModel(Model):
    @classmethod
    def get_one(cls, raise_not_found=True, to_dict=True, **kwargs):
        """

        :param raise_not_found:
        :param to_dict:
        :param kwargs:
        :return:
        """
        res = cls.query.filter_by(**kwargs)
        if raise_not_found:
            res = res.first_or_404()
        else:
            res = res.first()
            if res is None:
                return

        if to_dict is True:
            return res.to_dict()
        else:
            return res

    @classmethod
    def get_list(cls, to_dict=True, **kwargs):
        """

        :param to_dict:
        :param kwargs:
        :return:
        """
        res = cls.query.filter_by(**kwargs).all()
        if to_dict is not True:
            return res

        return [r.to_dict() for r in res]



from datetime import datetime

import flask
from flask_ipban import IpBan as _IPBan
from flask_limiter import Limiter

from flaskel import cap, httpcode
from flaskel.ext import cfremote
from flaskel.utils.datetime import Day


def response_ok(r):
    return httpcode.is_ok(r.status_code)


def response_ko(r):
    return r.status_code != httpcode.TOO_MANY_REQUESTS \
           and httpcode.is_ko(r.status_code)


limiter = Limiter(
    key_func=cfremote.get_remote,
    default_limits=[lambda: cap.config.LIMITER.FAIL],
    default_limits_deduct_when=response_ko
)


def header_whitelist():
    hdr = cap.config.LIMITER.BYPASS_KEY
    value = cap.config.LIMITER.BYPASS_VALUE
    return flask.request.headers.get(hdr) == value


@limiter.request_filter
def _header_whitelist():
    return header_whitelist()


class RateLimit:
    limiter = limiter

    @classmethod
    def slow(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.SLOW,
            deduct_when=response_ok
        )

    @classmethod
    def medium(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.MEDIUM,
            deduct_when=response_ok
        )

    @classmethod
    def fast(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.FAST,
            deduct_when=response_ok
        )

    @classmethod
    def fail(cls):
        return cls.limiter.limit(
            lambda: cap.config.LIMITER.FAIL,
            deduct_when=response_ko
        )


class IPBan(_IPBan):
    def _after_request(self, response):
        """
        method to call after a request to allow recording all unwanted status codes

        :param response:
        :return:
        """
        statuses = (httpcode.NOT_FOUND, httpcode.METHOD_NOT_ALLOWED, httpcode.NOT_IMPLEMENTED)
        if response.status_code in statuses:
            self.add()

        return response

    def _before_request_check(self):
        self.ip_record.update_from_other_instances()
        if self._is_excluded():
            return

        ip = self.get_ip()
        entry = self._ip_ban_list.get(ip)
        url = flask.request.environ.full_path

        if entry and entry.get('count', 0) > self.ban_count:
            if entry.get('permanent', False):
                flask.abort(httpcode.GONE)

            now = datetime.now()
            delta = now - entry.get('timestamp', now)
            if delta.seconds < self.ban_seconds or self.ban_seconds == 0:
                self._logger.info(f"IP: {ip} updated in ban list, at url: {url}")
                entry['count'] += 1
                entry['timestamp'] = now
                self.ip_record.write(ip, count=entry['count'])
                flask.abort(httpcode.GONE)

            entry['count'] = 0
            self._logger.debug(f"IP: {ip} expired from ban list, at url: {url}")

    def load_nuisances(self, conf=None):
        """
        load a yaml file of nuisance urls or ips that are commonly used by vulnerability scanners
        any access to one of these that produces unwanted status codes will ban the source ip.
        Each call to load_nuisances will add to the current list of nuisances

        :param conf:
        :return: the number of nuisances added from this file
        """
        count = 0
        for match_type in ['ip', 'string', 'regex']:
            values = conf.get(match_type) or ()
            for v in values:
                try:
                    self.url_block_pattern_add(v, match_type)
                    count += 1
                except Exception as e:
                    self._logger.exception(f"Exception {e} adding pattern {v}")

        return count


class FlaskIPBan:
    def __init__(self, app=None, ban=None, **kwargs):
        """

        :param app:
        :param ban: optional IpBan instance
        """
        self.ip_ban = ban or IPBan()

        if app:
            self.init_app(app, **kwargs)

    def init_app(self, app, white_list=None, nuisances=None):
        """

        :param app:
        :param white_list:
        :param nuisances:
        """
        app.config.setdefault('IPBAN_COUNT', 20)
        app.config.setdefault('IPBAN_SECONDS', Day.seconds)
        self.ip_ban.ban_count = app.config.IPBAN_COUNT
        self.ip_ban.ban_seconds = app.config.IPBAN_SECONDS

        self.ip_ban.init_app(app)
        self.ip_ban.ip_whitelist_add(white_list or [])
        self.ip_ban.load_nuisances(conf=cap.config.IPBAN_NUISANCES or nuisances or {})



import functools
import inspect

from flask import request
from flask.views import View

import flaskel.http.rpc as rpc
from flaskel import cap, httpcode
from flaskel.ext import builder
from flaskel.utils.datastruct import ObjectDict


class JSONRPCView(View):
    version = '2.0'
    separator = '.'
    methods = ['POST']
    decorators = [builder.response('json')]

    def __init__(self, operations=None):
        """

        :param operations: a dict of jsonrpc operations like:
            {
                None: {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                "method1": {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                ...
            }
        """
        self.operations = operations or {}

    def _validate_payload(self):
        """

        :return:
        """
        payload = request.get_json(True)
        if not payload:
            raise rpc.RPCParseError() from None

        if 'jsonrpc' not in payload or \
                'method' not in payload:
            raise rpc.RPCInvalidRequest() from None

        if payload['jsonrpc'] != self.version:
            raise rpc.RPCInvalidRequest(f"jsonrpc version is {self.version}") from None

        return payload

    def _get_action(self, method):
        """

        :param method:
        :return:
        """
        m = method.split(self.separator)
        if len(m) > 1:
            op, action = m[0], self.separator.join(m[1:])
        else:
            op, action = None, m[0]

        try:
            return self.operations[op][action]
        except (IndexError, TypeError, KeyError) as exc:
            cap.logger.debug(exc)
            raise rpc.RPCMethodNotFound()

    def dispatch_request(self):
        """

        :return:
        """
        response = ObjectDict(jsonrpc=self.version, id=None)

        try:
            payload = self._validate_payload()
            response.id = payload.get('id')

            action = self._get_action(payload['method'])
            response.result = action(**(payload.get('params') or {}))

            if 'id' not in payload:
                return None, httpcode.NO_CONTENT

        except rpc.RPCError as ex:
            response.error = ex.as_dict()
        except Exception as ex:
            cap.logger.exception(ex)
            mess = str(ex) if cap.debug is True else None
            response.error = rpc.RPCInternalError(message=mess).as_dict()

        return response

    def load_from_object(self, obj):
        """

        :param obj:
        """
        for m in inspect.getmembers(obj, predicate=inspect.isroutine):
            if not m[0].startswith('_'):
                self.method(obj.__class__.__name__, m[0])(m[1])

    def method(self, name=None, operation=None):
        """

        :param name:
        :param operation:
        :return:
        """

        def _method(func):
            """

            :param func:
            :return:
            """
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                obj = {operation: func}
                if name not in self.operations:
                    self.operations[name] = obj
                else:
                    self.operations[name].update(obj)

            return wrapped()

        return _method

    def register(self, app, name=None, url=None, **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name: view name
        :param url: url to bind
        """
        name = name or self.__class__.__name__
        url = url or f"/{name}"
        view_func = self.__class__.as_view(name, self.operations, **kwargs)
        app.add_url_rule(f"/{url.rstrip('/')}", view_func=view_func, methods=self.methods)



# noinspection PyUnresolvedReferences
from flaskel.tasks import celery

broker_url = "redis://localhost:6379/0"
result_backend = "redis://localhost:6379/0"
include = ["tasks"]

broker_transport_options = {
    "socket_timeout": 5,
    "max_retries":    4,
    "interval_start": 0,
    "interval_step":  0.5,
    "interval_max":   3
}

broker_connection_max_retry = 3



import flask
from sqlalchemy.exc import IntegrityError

from flaskel import httpcode
from .base import Resource


class CatalogResource(Resource):
    def __init__(self, model):
        self._model = model

    def on_get(self, res_id, *args, **kwargs):
        return self._model.get_one(id=res_id)

    def on_collection(self, *args, **kwargs):
        res = self._model.get_list(to_dict=False)
        restricted = False if flask.request.args.get('_related') else True
        return [r.to_dict(restricted=restricted) for r in res]


class Restful(CatalogResource):
    """TODO Not tested yet"""

    def __init__(self, db, model):
        """

        :param db:
        :param model:
        """
        super().__init__(model)
        self._db = db

    def on_post(self, payload=None, *args, **kwargs):
        """

        :param payload:
        :return:
        """
        res = self._model(**(payload or {}))
        try:
            self._db.session.add()
            self._db.session.commit()
        except IntegrityError:
            self._db.session.rollback()
            flask.abort(httpcode.CONFLICT)
        finally:
            self._db.session.close()

        return res.to_dict(), httpcode.CREATED

    def on_delete(self, res_id, *args, **kwargs):
        """

        :param res_id:
        """
        res = self._model.query.get_or_404(res_id)
        res.delete()
        self._db.session.commit()

    def on_put(self, res_id, payload=None, *args, **kwargs):
        """

        :param res_id:
        :param payload:
        :return:
        """
        res = self._model.query.get_or_404(res_id)
        res.update(**(payload or {}))
        self._db.session.add(res)
        self._db.session.commit()
        return res.to_dict()



import argon2 as _argon2
from argon2.exceptions import Argon2Error, InvalidHash


class Argon2:
    def __init__(self, app=None):
        """

        :param app:
        """
        self._ph = None

        if app is not None:
            self.init_app(app)  # pragma: no cover

    def init_app(self, app):
        """

        :param app:
        """
        app.config.setdefault("ARGON2_ENCODING", "utf-8")
        app.config.setdefault("ARGON2_TIME_COST", _argon2.DEFAULT_TIME_COST)
        app.config.setdefault("ARGON2_HASH_LENGTH", _argon2.DEFAULT_HASH_LENGTH)
        app.config.setdefault("ARGON2_MEMORY_COST", _argon2.DEFAULT_MEMORY_COST)
        app.config.setdefault("ARGON2_PARALLELISM", _argon2.DEFAULT_PARALLELISM)
        app.config.setdefault("ARGON2_RANDOM_SALT_LENGTH", _argon2.DEFAULT_RANDOM_SALT_LENGTH)

        self._ph = _argon2.PasswordHasher(
            encoding=app.config['ARGON2_ENCODING'],
            time_cost=app.config['ARGON2_TIME_COST'],
            hash_len=app.config['ARGON2_HASH_LENGTH'],
            memory_cost=app.config['ARGON2_MEMORY_COST'],
            parallelism=app.config['ARGON2_PARALLELISM'],
            salt_len=app.config['ARGON2_RANDOM_SALT_LENGTH']
        )

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['argon2'] = self

    def generate_hash(self, password):
        """

        :param password:
        :return:
        """
        return self._ph.hash(password)

    def verify_hash(self, pw_hash, password, exc=False):
        """

        :param pw_hash:
        :param password:
        :param exc:
        :return:
        """
        try:
            return self._ph.verify(pw_hash, password)
        except (Argon2Error, InvalidHash):
            if exc is True:
                raise  # pragma: no cover
            return False



from multiprocessing import cpu_count

from waitress import serve

from .base import BaseApplication


class WSGIWaitress(BaseApplication):
    def run(self):
        """

        :return:
        """
        self.options.setdefault('threads', cpu_count())

        serve(
            self.application,
            host=self._interface,
            port=self._port,
            threads=self.options['threads'],
        )



import importlib

from sqlalchemy import MetaData
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import class_mapper
from sqlalchemy_schemadisplay import create_schema_graph, create_uml_graph


def model_to_uml(module):
    """

    :param module:
    :return:
    """
    mappers = []
    models = importlib.import_module(module)

    for attr in dir(models):
        if attr[0] == '_':
            continue
        try:
            cls = getattr(models, attr)
            mappers.append(class_mapper(cls))
        except SQLAlchemyError:
            pass

    return create_uml_graph(
        mappers,
        show_operations=False
    )


def db_to_schema(url):
    """

    :param url:
    :return:
    """
    return create_schema_graph(
        metadata=MetaData(url),
        show_datatypes=False,
        show_indexes=False,
        rankdir='LR',
        concentrate=False
    )



import os

from decouple import config

DEBUG = config('DEBUG', default=False)
TESTING = config('TESTING', default=DEBUG)
APP_NAME = config('APP_NAME', default="flaskel")
APP_HOST = config('APP_HOST', default='127.0.0.1')
APP_PORT = config('APP_PORT', default=5000)
FLASK_APP = config('FLASK_APP', default="app:app")
FLASK_ENV = config('FLASK_ENV', default="production")
SERVER_NAME = config('SERVER_NAME', default=f"{APP_HOST}:{APP_PORT}")

LOCALE = config('LOCALE', default="en_EN.utf8")
TEMPLATES_AUTO_RELOAD = config("TEMPLATES_AUTO_RELOAD", default=DEBUG)
EXPLAIN_TEMPLATE_LOADING = config("EXPLAIN_TEMPLATE_LOADING", default=DEBUG)

BASE_DIR = config("BASE_DIR", default=os.path.abspath(os.path.dirname(__file__)))
LOG_FILE_CONF = config("LOG_FILE_CONF", default=os.path.join('config', 'log.yaml'))

SECRET_KEY_MIN_LENGTH = 256

REQUEST_METHODS = {
    'WITHOUT_BODY': [
        'GET', 'DELETE'
    ],
    'WITH_BODY':    [
        'POST', 'PUT', 'PATCH'
    ]
}

ALLOWED_METHODS = REQUEST_METHODS['WITH_BODY'] + REQUEST_METHODS['WITHOUT_BODY']

AC_ALLOW_ORIGIN = []

RB_DEFAULT_ACCEPTABLE_MIMETYPES = [
    "application/json",
    "application/xml"
]

ERROR_PAGE = 'core/error.html'
LOG_LOGGER_NAME = FLASK_ENV
LOG_APP_NAME = APP_NAME

CF_STRICT_ACCESS = False

USE_X_SENDFILE = True
SEND_FILE_MAX_AGE_DEFAULT = 86400
ENABLE_ACCEL = True
ACCEL_BUFFERING = True
ACCEL_CHARSET = 'utf-8'
ACCEL_LIMIT_RATE = 'off'

WSGI_WERKZEUG_LINT_ENABLED = True
WSGI_WERKZEUG_PROFILER_ENABLED = True
WSGI_WERKZEUG_PROFILER_RESTRICTION = (0.1,)
WSGI_WERKZEUG_PROFILER_FILE = 'profiler.txt'

DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"

REQUEST_ID_HEADER = 'X-Request-ID'
RATELIMIT_HEADERS_ENABLED = True
LIMITER = {
    'FAIL':         '1/second',
    'FAST':         '30/minute',
    'MEDIUM':       '20/minute',
    'SLOW':         '10/minute',
    'BYPASS_KEY':   'X-Limiter-Bypass',
    'BYPASS_VALUE': 'bypass-rate-limit',
}



from flask import current_app as cap


class HTTPDumper:
    @staticmethod
    def padding(text):
        """

        :param text:
        :return:
        """
        return "\n{}".format(text)

    @classmethod
    def dump_headers(cls, hdr, only=()):
        """
        dumps http headers: useful for logging

        :param hdr: headers' dictionary
        :param only: list of headers key to dump
        :return: string representation of headers
                k1: v1
                k2: v2
        """
        if only:
            hdr = {k: hdr[k] for k in only if k in hdr}

        return cls.padding('\n'.join('{}: {}'.format(k, v) for k, v in hdr.items()))

    @classmethod
    def response_filename(cls, headers):
        """
        get filename from Content-Disposition header.
        i.e.: attachment; filename="<file name.ext>"

        :param headers: http headers dict
        :return: only the file name
        """
        hdr = headers.get('Content-Disposition')
        if not hdr:
            return None  # pragma: no cover

        tmp = hdr.split(';')
        hdr = tmp[1] if len(tmp) > 1 else tmp[0]
        tmp = hdr.split('=')

        if len(tmp) > 1:
            return tmp[1].strip('"').lstrip('<').rstrip('>')
        else:
            return None

    @classmethod
    def dump_request(cls, req, dump_body=None, only_hdr=()):
        """
        dump http request: useful for logging

        :param req: Request instance
        :param dump_body: flag to enable or disable dump of request's body (overrides default)
        :param only_hdr: dump only a subset of headers
        :return: prettified representation of input as string
        """
        body = None
        if dump_body is True:
            try:
                body = req.body or "empty body"
                if type(body) is not str:
                    body = body.decode()
            except AttributeError as exc:
                body = str(exc)
        if body:
            body = f"body:\n{body}"

        headers = cls.dump_headers(req.headers, only=only_hdr)
        if headers:
            headers = f"headers:\n{headers}\n"

        return f"{req.method} {req.url}\n{headers}{body}"

    @classmethod
    def dump_response(cls, resp, dump_body=None, only_hdr=()):
        """
        dump http response: useful for logging

        :param resp: Response instance
        :param dump_body: flag to enable or disable dump of response's body (overrides default)
        :param only_hdr: dump only a subset of headers
        :return: prettified representation of input as string
        """
        body = None
        if dump_body is True:
            body = cls.response_filename(resp.headers) or resp.text
            try:
                if type(body) is not str:
                    body = body.decode()
            except AttributeError as exc:
                body = str(exc)

        try:
            seconds = resp.elapsed.total_seconds()
        except AttributeError:  # because aiohttp.ClientResponse has not elapsed attribute
            seconds = 'N/A'

        if body:
            body = f"body:\n{body}"

        headers = cls.dump_headers(resp.headers, only=only_hdr)
        if headers:
            headers = f"headers:\n{headers}\n"
        status = resp.status_code if hasattr(resp, 'status_code') else resp.status
        return f"time : {seconds} - status code: {status}\n{headers}{body}"


class FlaskelHTTPDumper:
    @classmethod
    def dump_request(cls, req, dump_body=None):
        return HTTPDumper.dump_request(
            req, dump_body, only_hdr=cap.config.LOG_REQ_HEADERS
        )

    @classmethod
    def dump_response(cls, resp, dump_body=None):
        return HTTPDumper.dump_response(
            resp, dump_body, only_hdr=cap.config.LOG_RESP_HEADERS
        )



from datetime import timedelta

import flask_jwt_extended as jwt
from flask_httpauth import HTTPBasicAuth

from flaskel import cap, httpcode
from .sqlalchemy import db

jwtm = jwt.JWTManager()
basic_auth = HTTPBasicAuth()


@basic_auth.verify_password
def simple_basic_auth(username, password):
    if username == cap.config.BASIC_AUTH_USERNAME \
            and password == cap.config.BASIC_AUTH_PASSWORD:
        return dict(username=username, password=password)


@jwtm.invalid_token_loader
def invalid_token_loader(mess):
    return dict(message=mess), httpcode.UNAUTHORIZED  # pragma: no cover


class RevokedTokenModel(db.Model):
    __tablename__ = 'revoked_tokens'

    id = db.Column(db.Integer, primary_key=True)
    jti = db.Column(db.String(120), nullable=False, unique=True)

    def __repr__(self):  # pragma: no cover
        return "<RevokedToken: %r>" % self.jti

    @classmethod
    def is_jti_blacklisted(cls, jti):  # pragma: no cover
        return bool(cls.query.filter_by(jti=jti).first())


@jwtm.token_in_blacklist_loader
def check_if_token_in_blacklist(decrypted_token):  # pragma: no cover
    return RevokedTokenModel.is_jti_blacklisted(decrypted_token['jti'])


def refresh_access_token(expires=None):
    """

    :param expires: in seconds
    :return:
    """
    access_token = jwt.create_access_token(
        identity=jwt.get_jwt_identity(),
        expires_delta=timedelta(expires) if expires else None
    )
    decoded = jwt.decode_token(access_token)
    return dict(
        access_token=access_token,
        expires_in=decoded['exp'],
        issued_at=decoded['iat'],
        token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
        scope=cap.config.JWT_DEFAULT_SCOPE
    )


def create_tokens(identity, expires_access=None, expires_refresh=None, scope=None):
    """

    :param identity: user identifier, generally the username
    :param expires_access: in seconds
    :param expires_refresh: in seconds
    :param scope:
    :return:
    """
    expires = timedelta(expires_access) if expires_access else None
    access_token = jwt.create_access_token(
        identity=identity, expires_delta=expires
    )

    expires = timedelta(expires_refresh) if expires_refresh else None
    refresh_token = jwt.create_refresh_token(
        identity=identity, expires_delta=expires
    )

    decoded = jwt.decode_token(access_token)
    expires_in = decoded['exp']
    issued_at = decoded['iat']

    return dict(
        access_token=access_token,
        refresh_token=refresh_token,
        expires_in=expires_in,
        issued_at=issued_at,
        token_type=cap.config.JWT_DEFAULT_TOKEN_TYPE,
        scope=scope or cap.config.JWT_DEFAULT_SCOPE
    )



import functools

import flask
from webargs import flaskparser

from flaskel import httpcode

parser = flaskparser.FlaskParser()
query = functools.partial(parser.use_args, location="query")
payload = functools.partial(parser.use_args, location="json")


# noinspection PyUnusedLocal
@parser.error_handler
def handle_error(error, *args, **kwargs):
    flask.abort(httpcode.BAD_REQUEST, response=error.messages)



import flask
from flask import current_app as cap
from webargs import fields

from flaskel import httpcode
from flaskel.ext.auth import basic_auth, create_tokens, jwt, refresh_access_token
from flaskel.utils.webargs import query
from . import auth


@auth.route('/access_token', methods=['POST'])
@basic_auth.login_required()
@query({
    'expire_access':  fields.Integer(missing=None, validate=lambda x: x > 0),
    'expire_refresh': fields.Integer(missing=None, validate=lambda x: x > 0),
})
def access_token(args):
    data = flask.request.json
    if data.email == 'email' and data.password == 'password':
        return create_tokens(
            data.email,
            expires_access=args['expire_access'],
            expires_refresh=args['expire_refresh'],
        )

    flask.abort(httpcode.UNAUTHORIZED)


@auth.route('/refresh_token', methods=['GET', 'POST'])
@jwt.jwt_refresh_token_required
def refresh_token():
    return refresh_access_token()


@auth.route('/check_token', methods=['GET'])
@jwt.jwt_required
def token_check():
    return dict(
        token_type=cap.config['JWT_DEFAULT_TOKEN_TYPE'],
        scope=cap.config['JWT_DEFAULT_SCOPE'],
        **jwt.get_raw_jwt()
    )



"""
Flaskel
-------
"""
import os
import re
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test

package = 'flaskel'
skeleton = 'skeleton'
BASE_PATH = os.path.dirname(__file__)
SKEL_DIR = os.path.join(BASE_PATH, skeleton)
VERSION_FILE = os.path.join(package, 'version.py')


def read(file):
    """

    :param file:
    :return:
    """
    with open(os.path.join(BASE_PATH, file)) as f:
        return f.read()


def grep(file, name):
    """

    :param file:
    :param name:
    :return:
    """
    pattern = r"{attr}\W*=\W*'([^']+)'".format(attr=name)
    value, = re.findall(pattern, read(file))
    return value


def readme(file):
    """

    :param file:
    :return:
    """
    try:
        return read(file)
    except OSError as exc:
        print(str(exc), file=sys.stderr)


class PyTest(test):
    def finalize_options(self):
        """

        """
        test.finalize_options(self)

    def run_tests(self):
        """

        """
        import pytest
        sys.exit(pytest.main(['tests']))


def package_files(directory):
    """

    :param directory:
    :return:
    """
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join(path, filename))

    return paths


try:
    os.symlink(SKEL_DIR, os.path.join(package, skeleton))

    setup(
        name='Flaskel',
        license='MIT',
        version=grep(VERSION_FILE, '__version__'),
        url='https://github.com/cs91chris/flaskel',
        author=grep(VERSION_FILE, '__author_name__'),
        author_email=grep(VERSION_FILE, '__author_email__'),
        description='Skeleton for flask applications',
        long_description=readme('README.rst'),
        platforms='any',
        python_requires='>3.6',
        zip_safe=False,
        packages=find_packages(),
        include_package_data=True,
        package_data={
            package: package_files(skeleton)
        },
        entry_points={
            'console_scripts': [
                'flaskel = flaskel.cli:cli',
            ],
        },
        install_requires=[
            "Flask",
            "Flask-Cors",
            "Flask-ErrorsHandler",
            "Flask-ResponseBuilder",
            "Flask-TemplateSupport",
            "Flask-CloudflareRemote",
            "Flask-Logify",
            "Flask-SqlAlchemy",
            "Flask-JWT-Extended",
            "Flask-Limiter",
            "Flask-IPBan",
            "Flask-HTTPAuth",
            "sqlalchemy_schemadisplay",
            "PyYAML",
            "requests",
            "user-agents",
            "argon2-cffi",
            "jsonschema",
            "webargs",
            "python-decouple",
            "pytest==6.2.*",
            "pytest-cov==2.10.*",
        ],
        extras_requires={},
        cmdclass={'test': PyTest},
        test_suite='tests',
        classifiers=[
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'Topic :: Software Development :: Libraries :: Python Modules'
        ]
    )
finally:
    os.unlink(os.path.join(package, skeleton))



from flask_sqlalchemy import event
from sqlalchemy.ext.declarative import declared_attr
from flaskel.utils.datastruct import ObjectDict
from . import db


class StandardMixin:
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    created_at = db.Column(db.DateTime, server_default=db.func.now())
    updated_at = db.Column(db.DateTime, onupdate=db.func.now())


class CatalogMixin:
    @declared_attr
    def __table_args__(self):
        return db.UniqueConstraint('label', 'type_id'),

    id = db.Column(db.Integer, primary_key=True)
    label = db.Column(db.String(100))
    type_id = db.Column(db.Integer)
    description = db.Column(db.String(250))

    def __str__(self):
        return self.label

    def to_dict(self):
        return ObjectDict(
            id=self.id,
            label=self.label,
            description=self.description,
            typeId=self.type_id
        )


class LoaderMixin:
    values = ()

    # noinspection PyUnusedLocal
    @classmethod
    def load_values(cls, *args, **kwargs):
        for d in cls.values:
            # noinspection PyArgumentList
            db.session.add(cls(**d))
        db.session.commit()

    @classmethod
    def register_loader(cls):
        # noinspection PyUnresolvedReferences
        event.listen(cls.__table__, 'after_create', cls.load_values)



def health_sqlalchemy(db):
    """

    :param db:
    :return:
    """
    try:
        with db.engine.connect() as connection:
            connection.execute('SELECT 1')
    except Exception as exc:
        return False, str(exc)
    return True, None  # pragma: no cover


def health_mongo(db):
    """

    :param db:
    :return:
    """
    try:
        db.db.command('ping')
    except Exception as exc:
        return False, str(exc)
    return True, None  # pragma: no cover


def health_redis(db):
    """

    :param db:
    :return:
    """
    try:
        db.ping()
    except Exception as exc:
        return False, str(exc)
    return True, None  # pragma: no cover



from flask import Blueprint

from flaskel.ext import cors, errors
from flaskel.views import JSONRPCView
from tests.blueprints.api.resource import APIResource
from tests.blueprints.api.rpc import MyJsonRPC

api = Blueprint('api', __name__, subdomain='api')

cors.init_app(api)
errors.api_register(api)

APIResource.register(api, 'resource_api', '/resources')

jsonRPCView = JSONRPCView()
jsonRPCView.load_from_object(MyJsonRPC())
jsonRPCView.register(api, 'myJsonRPC', "/rpc")



from gevent.pywsgi import WSGIServer

from .base import BaseApplication


class WSGIGevent(BaseApplication, WSGIServer):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        BaseApplication.__init__(self, app, options)
        WSGIServer.__init__(self, (self._interface, self._port), self.application)

    def run(self):
        """

        """
        self.serve_forever()
